import * as mc from '@minecraft/server';
import * as db from './database.js';

const playerShift = { x: 0, y: 2, z: 0 };

var beforeItemUseOn = { record: {}, instrument: {}, jukebox: {} };
var selectedSlots = {};
var stopRecord = {};
const recordSpots = new db.database('recording');
const noteblocks = new db.database('noteblocks');
const records = new db.database('records');

const pitch = [0.5, 0.529732, 0.561231, 0.594604, 0.629961, 0.66742, 0.707107, 0.749154, 0.793701, 0.840896, 0.890899, 0.943874, 1.0, 1.059463, 1.122462, 1.189207, 1.259921, 1.33484, 1.414214, 1.498307, 1.587401, 1.681793, 1.781797, 1.887749, 2.0];
const notes = ['F# - Oct. 0', 'G - Oct. 0', 'G# - Oct. 0', 'A - Oct. 0', 'A# - Oct. 0', 'B - Oct. 0',
    'C - Oct. 1', 'C# - Oct. 1', 'D - Oct. 1', 'D# - Oct. 1', 'E - Oct. 1', 'F - Oct. 1',
    'F# - Oct. 1', 'G - Oct. 1', 'G# - Oct. 1', 'A - Oct. 1', 'A# - Oct. 1', 'B - Oct. 1',
    'C - Oct. 2', 'C# - Oct. 2', 'D - Oct. 2', 'D# - Oct. 2', 'E - Oct. 2', 'F - Oct. 2', 'F# - Oct. 2']
const colors = [[0.466667, 0.843137, 0.0], [0.584314, 0.752941, 0.0], [0.698039, 0.647059, 0.0], [0.8, 0.52549, 0.0], [0.886275, 0.396078, 0.0], [0.952941, 0.254902, 0.0], [0.988235, 0.117647, 0.0], [0.996078, 0.0, 0.058824], [0.968627, 0.0, 0.2], [0.909804, 0.0, 0.352941], [0.811765, 0.0, 0.513725], [0.682353, 0.0, 0.662745], [0.52549, 0.0, 0.8], [0.356863, 0.0, 0.905882], [0.176471, 0.0, 0.976471], [0.007843, 0.039216, 0.996078], [0.0, 0.215686, 0.964706], [0.0, 0.407843, 0.878431], [0.0, 0.603922, 0.737255], [0.0, 0.776471, 0.552941], [0.0, 0.913725, 0.345098], [0.0, 0.988235, 0.129412], [0.121569, 0.988235, 0.0], [0.34902, 0.909804, 0.0], [0.580392, 0.756863, 0.0]]
const hex = ['77D700', '95C000', 'B2A500', 'CC8600', 'E26500', 'F34100', 'FC1E00', 'FE000F', 'F70033', 'E8005A', 'CF0083', 'AE00A9', '8600CC', '5B00E7', '2D00F9', '020AFE', '0037F6', '0068E0', '009ABC', '00C68D', '00E958', '00FC21', '1FFC00', '59E800', '94C100']
const instruments = {
    "Bass": "note.bass",
    "Snare Drum": "note.snare",
    "Hihat": "note.hat",
    "Bass Drum": "note.bd",
    "Bells": "note.bell",
    "Flute": "note.flute",
    "Chimes": "note.chime",
    "Guitar": "note.guitar",
    "Xylophone": "note.xylophone",
    "Iron Xylophone": "note.iron_xylophone",
    "Cow Bell": "note.cow_bell",
    "Didgeridoo": "note.didgeridoo",
    "Bit": "note.bit",
    "Banjo": "note.banjo",
    "Pling": "note.pling",
    "Harp": "note.harp"
}

function packLocation(block) {
    return [block.x, block.y, block.z].join('.');
}

function playNote(dimension, instrument, note, location, shift = { x: 0, y: 0, z: 0 }) {
    dimension.runCommandAsync(`playsound ${instruments[instrument]} @a ${location.x} ${location.y} ${location.z} 1 ${pitch[note]}`)
    dimension.spawnParticle('minecraft:note_particle', new mc.Location(location.x + shift.x, location.y + shift.y, location.z + shift.z), new mc.MolangVariableMap().setColorRGBA('variable.note_color', new mc.Color(colors[note][0], colors[note][1], colors[note][2], 1)));
}

async function playRecord(record, dimension, location) {
    let music = new db.database(`record${records.get()[record]}`).get().music;
    let playId = packLocation(location);
    for (let note of music) {
        if (stopRecord[playId] === true) {
            delete stopRecord[playId];
            return;
        }
        await setTimeout(() => {
            playNote(dimension, Object.keys(instruments)[note[1]], note[0], location, { x: 0.5, y: 1.5, z: 0.5 })
        }, note[2])
    }
}

function setTimeout(callback, delay) {
    return new Promise((resolve) => {
        let timeout = mc.system.runSchedule(() => {
            mc.system.clearRunSchedule(timeout);
            callback();
            resolve(true);
        }, delay);
    })
}

function getInstrument(block) {
    const blockmap = [
        [block.hasTag('wood'), 'Bass'],
        [block.hasTag('sand') || block.hasTag('gravel') || block.typeId == 'minecraft:concrete_powder', 'Snare Drum'],
        [block.typeId.includes('glass') || block.typeId == 'minecraft:beacon' || block.typeId == 'minecraft:sea_lantern', 'Hihat'],
        [block.hasTag('stone'), 'Bass Drum'],
        [block.typeId == 'minecraft:gold_block', 'Bells'],
        [block.typeId == 'minecraft:clay', 'Flute'],
        [block.typeId == 'minecraft:packed_ice', 'Chimes'],
        [block.typeId == 'minecraft:wool', 'Guitar'],
        [block.typeId == 'minecraft:bone_block', 'Xylophone'],
        [block.typeId == 'minecraft:iron_block', 'Iron Xylophone'],
        [block.typeId == 'minecraft:soul_sand', 'Cow Bell'],
        [block.hasTag('pumpkin'), 'Didgeridoo'],
        [block.typeId == 'minecraft:emerald_block', 'Bit'],
        [block.typeId == 'minecraft:hay_block', 'Banjo'],
        [block.typeId == 'minecraft:glowstone', 'pling'],
        [true, 'Harp']
    ];
    for (let block of blockmap) {
        let condition = block[0];
        let instrument = block[1];
        if (condition) {
            return instrument;
        }
    }
}

//Recording management
mc.world.events.beforeItemUseOn.subscribe((evt) => {
    let block = evt.source.dimension.getBlock(evt.blockLocation);
    let permutation = block.permutation;
    let player = evt.source;
    let inventory = player.getComponent("inventory").container;
    let item = evt.item;
    if (beforeItemUseOn.record[player.name] > Date.now()) return;
    beforeItemUseOn.record[player.name] = Date.now() + 250;
    if (block.typeId != 'minemusic:music_table') return;
    let mode = permutation.getProperty('minemusic:mode');

    if (mode.value == 'idle') {
        if (item.typeId.startsWith(`minecraft:music_disc`)) {
            mode.value = 'ready';
            block.setPermutation(permutation);
            let songData = recordSpots.get();
            songData[packLocation(block)] = { author: player.name, name: item.nameTag ?? 'Unnamed', music: [], pause: 0, record: false };
            recordSpots.set(songData);
            inventory.setItem(player.selectedSlot, new mc.ItemStack(mc.MinecraftItemTypes.diamond, 0));
        } else if (item.typeId == 'minemusic:music_disc_custom') {
            let lore = item.getLore()
            if (lore[1] == '§rOriginal') {
                mode.value = 'share';
                block.setPermutation(permutation);
                let songData = recordSpots.get();
                let songInfo = lore[0].slice(2).split(' - ')
                songData[packLocation(block)] = { author: songInfo[0], name: songInfo[1], music: [], pause: 0, record: false };
                recordSpots.set(songData);
                inventory.setItem(player.selectedSlot, new mc.ItemStack(mc.MinecraftItemTypes.diamond, 0));
            }
        }
    } else if (mode.value == 'ready') {
        mode.value = 'record';
        block.setPermutation(permutation);
        let recordData = recordSpots.get();
        let songData = recordData[packLocation(block)];
        let songInfo = `${songData.author} - ${songData.name}`;
        songData.record = true;
        recordSpots.set(recordData);
        player.runCommandAsync(`titleraw @s actionbar { "rawtext": [ { "translate": "langcode.recording" }, { "text": " ${songInfo}" } ] }`)
    } else if (mode.value == 'record') {
        mode.value = 'share';
        block.setPermutation(permutation);
        let recordData = recordSpots.get()
        let songData = recordData[packLocation(block)];
        let songInfo = `${songData.author} - ${songData.name}`;
        let songId;
        let songMap = records.get();

        if (songInfo in songMap) {
            songId = songMap[songInfo];
        } else {
            songId = Object.keys(songMap).length;
            songMap[songInfo] = songId;
            records.set(songMap)
        }

        let songDb = new db.database(`record${songId}`);
        songDb.set({ music: songData.music });
        player.runCommandAsync(`titleraw @s actionbar { "rawtext": [ { "translate": "langcode.recorded" }, { "text": " ${songInfo}" } ] }`)
    } else if (mode.value == 'share') {
        let recordData = recordSpots.get()
        let songData = recordData[packLocation(block)];
        let songInfo = `${songData.author} - ${songData.name}`;
        if (item.typeId.startsWith(`minecraft:music_disc`)) {
            let disc = new mc.ItemStack(mc.ItemTypes.get('minemusic:music_disc_custom'), 1);
            disc.setLore([`§r${songInfo}`, '§rCopy']);
            inventory.setItem(player.selectedSlot, disc);
        } else {
            mode.value = 'idle';
            block.setPermutation(permutation);
            let disc = new mc.ItemStack(mc.ItemTypes.get('minemusic:music_disc_custom'), 1);
            disc.setLore([`§r${songInfo}`, '§rOriginal']);
            player.dimension.spawnItem(disc, new mc.Location(block.x, block.y + 1, block.z));
            delete recordData[packLocation(block)];
            recordSpots.set(recordData);
        }
    }
})

//Note handler
mc.world.events.tick.subscribe((evt) => {
    for (let player of mc.world.getPlayers()) {
        if (!(player.name in selectedSlots)) {
            selectedSlots[player.name] = player.selectedSlot;
        }
        if (player.selectedSlot != selectedSlots[player.name]) {
            selectedSlots[player.name] = player.selectedSlot;
            let inventory = player.getComponent("inventory").container;
            let item = inventory.getItem(player.selectedSlot)
            if (!item) continue;
            if (item.typeId == 'minemusic:portable_noteblock') {
                let note = item.getLore()[0] ?? 0;
                if (note !== 0) {
                    note = notes.indexOf(note);
                }
                let instrument = item.getLore()[1] ?? 'Harp';
                playNote(player.dimension, instrument, note, player.location, playerShift);

                let recording = recordSpots.get();
                let songs = Object.keys(recording);
                let update = false;
                if (songs.length == 0) continue;
                for (let song of songs) {
                    let songData = recording[song];
                    if (!(songData.record)) continue;
                    let location = song.split('.');
                    let distance = { x: Math.abs(player.location.x - location[0]), z: Math.abs(player.location.z - location[2]) }
                    if (distance.x < 64 || distance.z < 64) {
                        update = true;
                        let pause;
                        if (songData.pause == 0) {
                            pause = 0
                            songData.pause = evt.currentTick;
                        } else {
                            pause = evt.currentTick - songData.pause;
                            songData.pause = evt.currentTick;
                        }
                        songData.music.push([note, Object.keys(instruments).indexOf(instrument), pause])
                    }
                }
                if (update) {
                    recordSpots.set(recording)
                }
            }
        }
    }
})

//Portable Noteblock
mc.world.events.itemUse.subscribe((evt) => {
    let item = evt.item;
    let lore = item.getLore();
    let player = evt.source;

    if (beforeItemUseOn.instrument[player.name] > Date.now()) return;

    let inventory = player.getComponent("inventory").container;
    if (item.typeId == 'minemusic:portable_noteblock') {
        if (lore[1] == undefined) {
            item.setLore([notes[0], 'Harp']);
            inventory.setItem(player.selectedSlot, item);
            player.runCommandAsync(`title @s actionbar ${notes[0]}`);
            playNote(player.dimension, 'Harp', 0, player.location, playerShift);
            return true;
        }
        let note = notes.indexOf(lore[0]);
        if (!(player.isSneaking)) {
            if (note < notes.length - 1) {
                item.setLore([notes[note + 1], lore[1]]);
                inventory.setItem(player.selectedSlot, item);
                player.runCommandAsync(`title @s actionbar §a${notes[note + 1]}`);
                playNote(player.dimension, lore[1], note + 1, player.location, playerShift);
            }
        } else {
            if (note > 0) {
                item.setLore([notes[note - 1], lore[1]]);
                inventory.setItem(player.selectedSlot, item);
                player.runCommandAsync(`title @s actionbar §c${notes[note - 1]}`);
                playNote(player.dimension, lore[1], note - 1, player.location, playerShift);
            }
        }
    }
})

mc.world.events.beforeItemUseOn.subscribe((evt) => {
    let player = evt.source;
    if (beforeItemUseOn.instrument[player.name] > Date.now()) return;
    beforeItemUseOn.instrument[player.name] = Date.now() + 250;

    let block = player.dimension.getBlock(evt.blockLocation);
    let item = evt.item;
    let lore = item.getLore();
    let inventory = player.getComponent("inventory").container;
    if (item.typeId == 'minemusic:portable_noteblock') {
        let instrument = getInstrument(block);
        if (lore[1] == undefined) {
            item.setLore([notes[0], instrument]);
            inventory.setItem(player.selectedSlot, item);
            player.runCommandAsync(`title @s actionbar ${instrument}`);
            playNote(player.dimension, instrument, 0, player.location, playerShift);
            return true;
        }
        let note = lore[0] ?? 0;
        if (note !== 0) {
            note = notes.indexOf(note);
        }
        item.setLore([lore[0], instrument]);
        inventory.setItem(player.selectedSlot, item);
        player.runCommandAsync(`title @s actionbar ${instrument}`);
        playNote(player.dimension, instrument, note, player.location, playerShift);
    }
})

mc.world.events.beforeItemUseOn.subscribe((evt) => {
    let player = evt.source;

    if (beforeItemUseOn.jukebox[player.name] > Date.now()) return;
    beforeItemUseOn.jukebox[player.name] = Date.now() + 500;

    let inventory = player.getComponent("inventory").container;
    let block = player.dimension.getBlock(evt.blockLocation);
    let item = evt.item;
    let itemId = item.typeId ?? 'minecraft:air'

    if (block.typeId == 'minecraft:jukebox') {
        let jb = noteblocks.get();
        let id = packLocation(block);
        if (jb[id] != undefined) {
            if (jb[id] !== 'vanilla') {
                let disc = new mc.ItemStack(mc.ItemTypes.get('minemusic:music_disc_custom'), 1);
                disc.setLore(jb[id]);
                player.dimension.spawnItem(disc, new mc.Location(block.x, block.y + 1, block.z));
                stopRecord[packLocation(block)] = true;
            }
            delete jb[id];
            noteblocks.set(jb);
        } else {
            if (itemId.startsWith('minecraft:music_disc')) {
                jb[id] = 'vanilla';
            } else if (itemId == 'minemusic:music_disc_custom') {
                jb[id] = item.getLore();
                let songInfo = item.getLore()[0].slice(2);
                stopRecord[packLocation(block)] = false;
                playRecord(songInfo, block.dimension, block.location);
                inventory.setItem(player.selectedSlot, new mc.ItemStack(mc.MinecraftItemTypes.diamond, 0));
                player.runCommandAsync(`titleraw @s actionbar { "rawtext": [ { "translate": "langcode.playing" }, { "text": " ${songInfo}" } ] }`)
            }
            noteblocks.set(jb);
        }
    }
})